import tempfile
from functools import cached_property
from typing import Any
from unittest import TestCase
from unittest.mock import call, mock_open, patch, sentinel

import yaml
from testfixtures import LogCapture

from setapp.settings import BaseApplication, BaseSettings, SettingsConfigDict
from setapp.utils import override_environ


class EnvironTestSettings(BaseSettings):
    """Simple test settings with config_environ set."""

    model_config = SettingsConfigDict(config_environ="SETAPP_TEST_CONFIG")


class ConfigFilesTestSettings(BaseSettings):
    """Simple test settings with config_files set."""

    model_config = SettingsConfigDict(config_files=("~/.fred/setapp.conf", "/etc/fred/setapp.conf"))


class SettingsTest(TestCase):
    home_call = call("/home/holly/.fred/setapp.conf")
    system_call = call("/etc/fred/setapp.conf")

    def setUp(self):
        self.log_handler = LogCapture("setapp", propagate=False)

    def tearDown(self):
        self.log_handler.uninstall()

    def assertDefaults(self, settings: BaseSettings) -> None:
        """Check settings contain all default settings."""
        self.assertEqual(settings.model_dump(), BaseSettings().model_dump())

    def test_config_file_environ(self):
        config_file = tempfile.NamedTemporaryFile()
        with open(config_file.name, "w") as config_io:
            yaml.dump({"logging": {"secret": "Gazpacho!"}}, config_io)

        with override_environ(SETAPP_TEST_CONFIG=config_file.name):
            settings = EnvironTestSettings()

        self.assertEqual(settings.logging, {"secret": "Gazpacho!"})

    def test_config_file_environ_unknown(self):
        with override_environ(SETAPP_TEST_CONFIG="/does/not/exist"):
            with self.assertRaisesRegex(FileNotFoundError, "/does/not/exist"):
                EnvironTestSettings()

    def test_config_file_argv(self):
        config_file = tempfile.NamedTemporaryFile()
        with open(config_file.name, "w") as config_io:
            yaml.dump({"logging": {"secret": "Wine cellar"}}, config_io)

        with override_environ():
            settings = BaseSettings(config_file=config_file.name)

        self.assertEqual(settings.logging, {"secret": "Wine cellar"})

    def test_config_file_argv_unknown(self):
        with override_environ():
            with self.assertRaisesRegex(FileNotFoundError, "/does/not/exist"):
                BaseSettings(config_file="/does/not/exist")

    def test_config_file_home(self):
        data = yaml.dump({"logging": {"secret": "Rusty gate"}})

        with override_environ(HOME="/home/holly"):
            with patch("os.path.isfile", return_value=True):
                with patch("builtins.open", mock_open(read_data=data)) as open_mock:
                    settings = ConfigFilesTestSettings()

        self.assertEqual(settings.logging, {"secret": "Rusty gate"})
        self.assertIn(self.home_call, open_mock.mock_calls)
        self.assertNotIn(self.system_call, open_mock.mock_calls)

    def test_config_file_system(self):
        data = yaml.dump({"logging": {"secret": "Androids"}})

        with override_environ(HOME="/home/holly"):
            with patch("os.path.isfile", side_effect=[False, True]):
                with patch("builtins.open", mock_open(read_data=data)) as open_mock:
                    settings = ConfigFilesTestSettings()

        self.assertEqual(settings.logging, {"secret": "Androids"})
        self.assertNotIn(self.home_call, open_mock.mock_calls)
        self.assertIn(self.system_call, open_mock.mock_calls)

    def test_config_file_no_settings(self):
        with override_environ(HOME="/home/holly"):
            with patch("os.path.isfile", return_value=False):
                with patch("builtins.open", side_effect=FileNotFoundError) as open_mock:
                    settings = ConfigFilesTestSettings()

        self.assertDefaults(settings)
        self.assertNotIn(self.home_call, open_mock.mock_calls)
        self.assertNotIn(self.system_call, open_mock.mock_calls)
        # Check no logs were made.
        self.log_handler.check()

    def test_config_file_error(self):
        with override_environ(HOME="/home/holly"):
            with patch("builtins.open", side_effect=OSError) as open_mock:
                settings = ConfigFilesTestSettings()

        self.assertDefaults(settings)
        self.assertNotIn(self.home_call, open_mock.mock_calls)
        self.assertNotIn(self.system_call, open_mock.mock_calls)


class ApplicationTest(TestCase):
    def setUp(self):
        self.log_handler = LogCapture("setapp", propagate=False)

    def tearDown(self):
        self.log_handler.uninstall()

    def test_init(self):
        class TestApplication(BaseApplication):
            settings_cls = sentinel.default_cls

        data = (
            (TestApplication(), sentinel.default_cls),
            (TestApplication(sentinel.settings_cls), sentinel.settings_cls),
        )
        for application, settings_cls in data:
            with self.subTest(application=application):
                self.assertEqual(application.settings_cls, settings_cls)

    def test_init_unset(self):
        with self.assertRaisesRegex(ValueError, "Settings class unset"):
            BaseApplication()

    def test_settings(self):
        application = BaseApplication(BaseSettings)
        application._settings = sentinel.settings
        self.assertEqual(application.settings, sentinel.settings)

    def test_settings_empty(self):
        application = BaseApplication(BaseSettings)
        with self.assertRaisesRegex(RuntimeError, "Application is not loaded"):
            application.settings

    def test_settings_setter(self):
        application = BaseApplication(BaseSettings)
        application.settings = sentinel.settings
        self.assertEqual(application.settings, sentinel.settings)

    def test_settings_setter_nonempty(self):
        application = BaseApplication(BaseSettings)
        application._settings = sentinel.old_settings

        application.settings = sentinel.settings
        self.assertEqual(application.settings, sentinel.settings)

    def test_is_loaded(self):
        application = BaseApplication(BaseSettings)
        application._settings = sentinel.settings
        self.assertTrue(application.is_loaded)

    def test_is_loaded_empty(self):
        self.assertFalse(BaseApplication(BaseSettings).is_loaded)

    def test_load(self):
        application = BaseApplication(BaseSettings)

        application.load()

        self.assertIsNotNone(application.settings)

    def test_load_settings_kwargs(self):
        application = BaseApplication(BaseSettings)

        application.load(settings_kwargs={"sentry": {"dsn": "http://example.com/"}})

        self.assertEqual(application.settings.sentry.dsn, "http://example.com/")  # type: ignore[union-attr]

    def test_load_environ(self):
        config = {"logging": {"secret": "Gazpacho!"}}
        config_file = tempfile.NamedTemporaryFile()
        with open(config_file.name, "w") as config_io:
            yaml.dump(config, config_io)

        application = BaseApplication(EnvironTestSettings)
        with patch("setapp.settings.dictConfig"):
            with override_environ(SETAPP_TEST_CONFIG=config_file.name):
                application.load()

        self.assertEqual(application.settings.model_dump(), {"logging": {"secret": "Gazpacho!"}, "sentry": None})

    def test_load_file(self):
        config = {"logging": {"secret": "Gazpacho!"}}
        config_file = tempfile.NamedTemporaryFile()
        with open(config_file.name, "w") as config_io:
            yaml.dump(config, config_io)

        application = BaseApplication(BaseSettings)
        with patch("setapp.settings.dictConfig"):
            with override_environ():
                application.load(config_file.name)

        self.assertEqual(application.settings.model_dump(), {"logging": {"secret": "Gazpacho!"}, "sentry": None})

    def test_load_file_no_reload(self):
        # Test reload is not allowed.
        application = BaseApplication(BaseSettings)
        application.load()
        with self.assertRaisesRegex(RuntimeError, "Application already loaded"):
            application.load()

    def _test_load(self, config: dict[str, Any]) -> None:
        # Generic test of load method
        def _settings_cls(*args, **kwargs):
            return BaseSettings(**config)

        application = BaseApplication(_settings_cls)  # type: ignore[arg-type, var-annotated]
        application.load()

    def test_load_logging(self):
        log_config = {"secret": "Gazpacho!"}
        with patch("setapp.settings.dictConfig") as config_mock:
            self._test_load({"logging": log_config})

        self.assertEqual(config_mock.mock_calls, [call(log_config)])

    def test_load_no_sentry(self):
        with patch("setapp.settings.sentry_sdk") as sentry_mock:
            self._test_load({})

        self.assertEqual(sentry_mock.mock_calls, [])

    def test_load_sentry_missing(self):
        dsn = "SMEGHEAD!"
        with patch("setapp.settings.sentry_sdk", new=None):
            self._test_load({"sentry": {"dsn": dsn}})

        self.log_handler.check(("setapp.settings", "WARNING", "Sentry set up, but sentry-sdk not available."))

    def test_load_sentry(self):
        dsn = "SMEGHEAD!"
        environment = "JMC"
        release = "STAR-BUG"
        ca_certs = "Silver Swimming Certificate"
        data = (
            # settings, call
            ({"dsn": dsn}, call.init(dsn=dsn, environment=None, release=None, ca_certs=None)),
            (
                {"dsn": dsn, "environment": environment},
                call.init(dsn=dsn, environment=environment, release=None, ca_certs=None),
            ),
            ({"dsn": dsn, "release": release}, call.init(dsn=dsn, environment=None, release=release, ca_certs=None)),
            ({"dsn": dsn, "ca_certs": ca_certs}, call.init(dsn=dsn, environment=None, release=None, ca_certs=ca_certs)),
        )
        for settings, sentry_call in data:
            with self.subTest(settings=settings):
                with patch("setapp.settings.sentry_sdk") as sentry_mock:
                    self._test_load({"sentry": settings})

                self.assertEqual(sentry_mock.mock_calls, [sentry_call])

    def test_clear_cache_empty(self):
        values = [sentinel.first, sentinel.second]

        class TestApplication(BaseApplication):
            @cached_property
            def cached_value(self):
                return values.pop(0)

            def clear_cache(self):
                self._del("cached_value")

        application = TestApplication(sentinel.settings)

        # Clear cache
        application.clear_cache()

        # First access
        self.assertEqual(application.cached_value, sentinel.first)

    def test_clear_cache(self):
        values = [sentinel.first, sentinel.second]

        class TestApplication(BaseApplication):
            @cached_property
            def cached_value(self):
                return values.pop(0)

            def clear_cache(self):
                self._del("cached_value")

        application = TestApplication(sentinel.settings)
        # First access
        application.cached_value

        # Clear cache
        application.clear_cache()

        # Second access
        self.assertEqual(application.cached_value, sentinel.second)
